/* Copyright 2008-2014 Research Foundation State University of New York   */

/* This file is part of QUB Express.                                      */

/* QUB Express is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Express is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

/*
begin_html
<p>
 See also:
 <ul>
 </ul>
 <b>Up:</b> <a href="index.html">Index</a>
</p>
end_html
*/

#include "qub_kmeans.h"

#ifdef _WIN32
  #include <windows.h>
  #include <time.h>
  #include <process.h>
  #define sleep Sleep
  #define getpid _getpid
#else
  #include <stdlib.h>
  #include <unistd.h>
  #define BOOL int
  #define TRUE 1
  #define FALSE 0
#endif

#include <string.h>
#include <vector>
#include <map>
#include <iostream>

using std::cerr;
using std::endl;

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/variate_generator.hpp>

// boost::random:
typedef boost::mt19937 base_generator_type;
typedef boost::uniform_int<> uni_distribution_type;
typedef boost::variate_generator<base_generator_type&, uni_distribution_type> uni_gen_type;

// library lifetime to avoid duplicate seeds
//   func-lifetime generators with passed-in seed
//   simulate{1}() use global generator (maybe better named)
base_generator_type g_kmeans_randgen( static_cast<boost::mt19937::result_type>( clock() + (int)getpid() ) );



class qub_k_means
{
private:
  int N, K;
  std::vector<float> weights;
  int *colors;
  std::vector< std::vector<float> > series_stor;
  std::vector<float*> series;
  std::vector<double> means_stor, measures_stor;
  std::vector<double*> means, measures;
  std::vector<double> color_weight;
  std::vector<double> series_std;
  
  void alloc_k(int k)
  {
    K = k;
    if ( means_stor.size() < (series.size()*(k+1)) ) {
      means_stor.resize(series.size()*(k+1));
      means.resize(series.size());
      double *row = & means_stor[0];
      for (size_t i=0; i<series.size(); ++i, row += k+1)
	means[i] = row;
    }
    if ( ((int)measures_stor.size()) < (N*(k+1)) ) {
      measures_stor.resize(N*(k+1));
      measures.resize(N);
      double *row = & measures_stor[0];
      for (int i=0; i<N; ++i, row += k+1)
	measures[i] = row;
    }
    if ( ((int)color_weight.size()) < (k+1) )
      color_weight.resize(k+1);
  }
  
  void calc_color_weight()
  {
    for (int i=0; i<=K; ++i)
      color_weight[i] = 0.0;
    for (int i=0; i<N; ++i) {
      int c = colors[i];
      if ( c )
	color_weight[colors[i]] += weights[i];
    }
  }
  
  void calc_means()
  {
    for (size_t s=0; s<series.size(); ++s) {
      for (int c=0; c<=K; ++c) {
	double sum = 0.0;
	if ( c && color_weight[c] ) {
	  for (int i=0; i<N; ++i) {
	    if ( colors[i] == c )
	      sum += series[s][i] * weights[i];
	  }
	}
	if ( color_weight[c] )
	  means[s][c] = sum / color_weight[c];
	// cerr << "means[" << s << "][" << c << "] = " << means[s][c] << endl;
      }
    }
  }

  void calc_measures()
  {
    for (int i=0; i<N; ++i) {
      for (int c=0; c<=K; ++c) {
	double sum = 0.0;
	if ( c ) {
	  for (size_t s=0; s<series.size(); ++s) {
	    double diff = (series[s][i] - means[s][c]);
	    sum += diff*diff;
	  }
	} else {
	  sum = 1e10;
	}
	measures[i][c] = sum;
      }
    }
  }
  
  double calc_score(int *picked)
  {
    double score = 0.0;
    for (int i=0; i<N; ++i)
      score += measures[i][picked[i]] * weights[i];
    return score;
  }
  
  void random_colors()
  {
    uni_gen_type uniform(g_kmeans_randgen, uni_distribution_type(0, N-1));
    for (int i=0; i<N; ++i)
      colors[i] = 0;
    std::map<int, int> used;
    used[-1] = 1;
    for (int c=1; c<=K && c<N; ++c) {
      int pos = -1;
      while ( used.find(pos) != used.end() ) {
	pos = uniform();
      }
      used[pos] = 1;
      colors[pos] = c;
    }
  }
  
  void pick_colors(int *picked)
  {
    for (int i=0; i<N; ++i) {
      double lo = measures[i][1];
      int lo_c = 1;
      for (int c=2; c<=K; ++c) {
	if ( measures[i][c] < lo ) {
	  lo = measures[i][c];
	  lo_c = c;
	}
      }
      picked[i] = lo_c;
    }
  }
  
  void set_colors(int *picked)
  { // and canonically renumber them
    std::map<int, int> renumber;
    int nextNum = 1;
    for (int i=0; i<N; ++i)
      if ( renumber.find(picked[i]) == renumber.end() )
	renumber[picked[i]] = nextNum++;
    for (int i=0; i<N; ++i)
      colors[i] = renumber[picked[i]];
  }
  
public:
  qub_k_means(int N_, float *weights_)
    : N(N_), weights(N_)
  {
    memcpy(&weights[0], weights_, N_*sizeof(float));
  }
    
  void add_series(float *series_)
  {
    int iser = (int) series_stor.size();
    series_stor.resize(iser+1);
    series_stor[iser].resize(N);
    series.resize(iser+1);
    for (size_t i=0; i<series.size(); ++i)
      series[i] = &(series_stor[i][0]);
    memcpy(series[iser], series_, N*sizeof(float));
  }
  
  double cluster(int *colors_)
  {
    colors = colors_;
    int k = 1;
    bool any_uncolored = false;
    for (int i=0; i<N; ++i) {
      if ( colors_[i] > k )
	k = colors_[i];
      else if ( colors[i] == 0 )
	any_uncolored = true;
    }
    alloc_k(k);
    
    calc_color_weight();
    calc_means();
    calc_measures();
    double last_ssd = 1e20;
    if ( ! any_uncolored ) {
      last_ssd = calc_score(colors_);
    }
    std::vector<int> picks(N);
    std::vector<int> best_picks(N);
    memcpy(&best_picks[0], colors_, N*sizeof(int));
    double best_ssd = last_ssd;
    int uphill_count = 0;
    for (int t=0; t<100*N*k; ++t) {
      pick_colors(&picks[0]);
      double ssd = calc_score(&picks[0]);
      if ( ssd >= last_ssd ) {           
	// cerr << "uphill" << endl;
	uphill_count += 1;                                                                                 
	if ( uphill_count > 2 )                                                                            
	  break;                                                                                         
      } else {                                                                                               
	uphill_count = 0;                                                                                  
      }
      set_colors(&picks[0]);
      calc_color_weight();
      calc_means();
      calc_measures();
      last_ssd = ssd;
      // cerr << "last_ssd: " << ssd << endl;
      if ( last_ssd < best_ssd ) {
	// cerr << "  is best_ssd" << endl;
	best_ssd = last_ssd;
	memcpy(&best_picks[0], &picks[0], N*sizeof(int));
      }
    }
    set_colors(&best_picks[0]);
    // cerr << k << ": " << best_ssd << endl;
    return best_ssd;
  }
    
  double cluster_trials(int *colors_, int k, int trials_per_k)
  {
    colors = colors_;
    alloc_k(k);
    double best_ssd = 1e20;
    std::vector<int> best_colors(N);
    memcpy(&best_colors[0], colors_, N*sizeof(int));
    if ( k < 2 )
      trials_per_k = 1;
    for (int trial=0; trial<trials_per_k; ++trial) {
      random_colors();
      double last_ssd = cluster(colors_);
      if ( last_ssd < best_ssd ) {
	best_ssd = last_ssd;
	memcpy(&best_colors[0], colors_, N*sizeof(int));
      }
    }
    memcpy(colors_, &best_colors[0], N*sizeof(int));
	return best_ssd;
  }
};

extern "C" QUBFAST_API void* qub_kmeans_create(int N, float *weights)
{
  return new qub_k_means(N, weights);
}

extern "C" QUBFAST_API void qub_kmeans_free(void *km)
{
  delete (qub_k_means*) km;
}

extern "C" QUBFAST_API void qub_kmeans_add_series(void *km, float *series)
{
  ((qub_k_means *) km)->add_series(series);
}

extern "C" QUBFAST_API double qub_kmeans_cluster(void *km, int *colors)
{
  return ((qub_k_means *) km)->cluster(colors);
}

extern "C" QUBFAST_API double qub_kmeans_cluster_trials(void *km, int *colors, int k, int trials_per_k)
{
  return ((qub_k_means *) km)->cluster_trials(colors, k, trials_per_k);
}

